// (c) 2015, SAS Institute Inc.
sap.ui.define([
    "jquery.sap.global",
    "sas/hc/m/Button",
    "./_FavoritesButton",
    "sas/hc/m/Label",
    "sas/hc/m/ResponsivePopover",
    "./_IconTabBar",
    "sas/hc/m/IconTabFilter",
    "./_ObjectInspectorPopoverContainer",
    "./_ButtonOverflowLayout",
    "./_SimplePairTable",
    "sas/hc/m/Toolbar",
    "sas/hc/ui/core/Icon"
], function(jQuery, Button, FavoritesButton, Label, ResponsivePopover, IconTabBar, IconTabFilter, ObjectInspectorPopoverContainer, ButtonOverflowLayout, SimplePairTable, Toolbar, Icon) {
    "use strict";

    /**
     * Constructor for a new ObjectInspectorBase.
     *
     * Base-class for ObjectInspector
     *
     * The aggregations defined in this control are rendered by the
     * ObjectInspectorPopoverContainer and ButtonOverflowLayout controls,
     * thus they are re-parented prior to rendering.
     *
     * For example, the actionButtons aggregation will be empty because the buttons
     * it is given are made aggregates of the ButtonOverflowLayout control or
     * it's ActionButton aggregation (depending on rendering logic).
     *
     * It is convienient for the consumer to have these aggregation defined at the
     * ObjectInspectorBase level for their binding and xml views, so there needs
     * needs to be a pass-through-like API defined at this level for the other
     * ManagedObject-level functions to work (ex: removeAllAggregations)
     *
     * @class
     * @name sas.hc.m.objectInspector.ObjectInspectorBase
     * @extends sas.hc.m.ResponsivePopover
     * @version 207001.0.1.20170511082246_f0htmcm27p
     * @author Jonathan Brink, Ralph Marshall
     * @constructor
     * @public
     */
    var ObjectInspectorBase = ResponsivePopover.extend("sas.hc.m.objectInspector.ObjectInspectorBase", /** @lends sas.hc.m.objectInspector.ObjectInspectorBase.prototype */ {
        metadata: {
            library: "sas.hc.m",
            properties: {
                headerText: {type: "string", defaultValue: "", group: "Data"},
                secondaryText: {type: "string", defaultValue: "", group: "Data"},
                description: {type: "string", defaultValue: "", group: "Data"},
                iconKey: {type: "string", defaultValue: "", group: "Data"},
                contentWidth: {type: "sap.ui.core.CSSSize", group: "Dimension"},
                contentHeight: {type: "sap.ui.core.CSSSize", group: "Dimension"},
                showOpenButton: {type: "boolean", defaultValue: true, group: "Behavior"},
                showEditButton: {type: "boolean", defaultValue: true, group: "Behavior"},
                showFavoritesButton: {type: "boolean", defaultValue: true, group: "Behavior"},
                objectData: {type: "object", group: "Data"},
                favoriteButtonSelectedDefault: {type: "boolean", defaultValue: false, group: "Behavior"},
                autoHeight: {type: "boolean", defaultValue: true, group: "Behavior"},
                objectDataLineHeight: {type: "integer", defaultValue: 5, group: "Appearance"},

                /*
                 * The sap.m.ResponsivePopover default is "Right"
                 * @override
                 */
                placement: {type: "sap.m.PlacementType", group: "Misc", defaultValue: sap.m.PlacementType.Vertical},

                /* @deprecated Since version 3.0 */
                maxAutoHeightObjectDataLines: {type: "integer", defaultValue: 10, group: "Appearance"}
            },
            aggregations: {
                details: {type: "sas.hc.m.IconTabFilter", multiple: true, singularName: "detail"},

                // rendered by ButtonOverflowLayoutToolbar, but still owned by ObjectInspector
                openButton: {type: "sas.hc.m.Button", multiple: false},
                editButton: {type: "sas.hc.m.Button", multiple: false},
                favoritesButton: {type: "sas.hc.m.objectInspector.FavoritesButton", multiple: false},
                actionButtons: {type: "sas.hc.m.Button", multiple: true, singularName: "actionButton"}
            },
            associations: {
                // added to content aggregation
                objectInspectorPopoverContainer: {type: "sas.hc.m.objectInspector.ObjectInspectorPopoverContainer"},
                buttonOverflowLayout: {type: "sas.hc.m.objectInspector.ButtonOverflowLayout"},

                // added to footer aggregation
                buttonOverflowLayoutToolbar: {type: "sas.hc.m.Toolbar"},

                // added to ObjectInspectorPopoverContainer's content aggregation
                iconTabBar: {type: "sas.hc.m.objectInspector.ObjectInspectorPopoverContainer"},

                // added as a detail to the IconTabBar
                simplePairTable: {type: "sas.hc.m.objectInspector.SimplePairTable"},

                // owned by ObjectInspectorPopoverContainer
                headerTextLabel: {type: "sas.hc.m.Label", multiple: false},
                secondaryTextLabel: {type: "sas.hc.m.Label", multiple: false},
                descriptionLabel: {type: "sas.hc.m.Label", multiple: false},
                headerIcon: {type: "sas.hc.ui.core.Icon", multiple: false}
            },
            events: {
                openButtonPress: {},
                editButtonPress: {},
                favoritesButtonPress: {}
            }
        },
        renderer: "sas.hc.m.PopoverRenderer",

        // resource bundle for ObjectInspector
        rb: sap.ui.getCore().getLibraryResourceBundle("sas.hc.m")
    });


    // =============================================================================
    // lifecycle functions
    // =============================================================================

    /**
     * init lifecycle callback.
     * Creates high-level controls that make up ObjectInspectorBase
     * @private
     */
    ObjectInspectorBase.prototype.init = function() {
        var self = this,
            sId = this.getId(),

            // objectInspectorPopoverContainer sub-controls
            oHeaderTextLabel = new Label(sId + "-HeaderTextLabel"),
            oSecondaryTextLabel = new Label(sId + "-SecondaryTextLabel"),
            oDescriptionLabel = new Label(sId + "-DescriptionLabel"),
            oHeaderIcon = new Icon(sId + "-HeaderIcon"),

            // holds all the popover content
            oObjectInspectorPopoverContainer = new ObjectInspectorPopoverContainer(sId + "-ObjectInspectorPopoverContainer"),

            // holds "details" aggregation.
            oIconTabBar = new IconTabBar({
                id: sId + "-DetailsIconTabBar",
                items: [],
                select: function(e) {
                    setTimeout(function() {
                        self._manageAutoSize();
                    }, 100);
                }
            }),

            // footer content (named and action buttons)
            oButtonOverflowLayout = new ButtonOverflowLayout({
                id: sId + "-ButtonOverflowLayout",
                objectInspector: this
            }),
            // footer container
            oButtonOverflowLayoutToolbar = new Toolbar({
                content: [
                    oButtonOverflowLayout
                ]
            }),

            // the "named" buttons
            oOpenButton = new Button({
                id: sId + '-NameButton_open',
                text: this.rb.getText("objectInspector.open.txt"),
                press: function(e) {
                    self.fireOpenButtonPress(e);
                }
            }),
            oEditButton = new Button({
                id: sId + '-NameButton_edit',
                text: this.rb.getText("objectInspector.edit.txt"),
                press: function(e) {
                    self.fireEditButtonPress(e);
                }
            }),
            oFavoritesButton = new FavoritesButton({
                id: sId + '-NameButton_favorites',
                press: function(e) {
                    self.fireFavoritesButtonPress(e);
                },
                selected: !!this.getFavoriteButtonSelectedDefault()
            });

        // invoke super
        if (ResponsivePopover.prototype.init) {
            ResponsivePopover.prototype.init.apply(this, arguments);
        }

        // force placement property to be respected
        this.setPlacement(this.getPlacement());

        // populate oObjectInspectorPopoverContainer
        sap.ui.getCore().byId(oObjectInspectorPopoverContainer.getHeaderTextContainer()).setMainContent(oHeaderTextLabel);
        sap.ui.getCore().byId(oObjectInspectorPopoverContainer.getSecondaryTextContainer()).setMainContent(oSecondaryTextLabel);
        sap.ui.getCore().byId(oObjectInspectorPopoverContainer.getDescriptionContainer()).setMainContent(oDescriptionLabel);
        sap.ui.getCore().byId(oObjectInspectorPopoverContainer.getHeaderIconContainer()).setMainContent(oHeaderIcon);

        // We have to add all of our locally implemenmted properties to this array so
        // that the code in the SAP version of ResponsiePopover does not attempt to
        // pass the function calls on to the underlying Popover/Dialog
        Array.prototype.push.apply(this._aNotSupportedProperties, [
            "headerText", "secondaryText", "description", "iconKey",
            "showOpenButton", "showEditButton",
            "showFavoritesButton", "objectData", "favoriteButtonSelectedDefault",
            "autoHeight", "objectDataLineHeight", "maxAutoHeightObjectDataLines"
        ]);

        // set defaults
        this.setProperty("showHeader", false, true);
        this.setModal(false);
        this.setAutoClose(true);
        this.setProperty("objectData", [], true);

        // S1205140: tell the popover/dialog what to give initial focus to
        this.setInitialFocus(oObjectInspectorPopoverContainer);

        // set aggregations/associations
        this.setHeaderTextLabel(oHeaderTextLabel);
        this.setSecondaryTextLabel(oSecondaryTextLabel);
        this.setDescriptionLabel(oDescriptionLabel);
        this.setHeaderIcon(oHeaderIcon);
        this.setIconTabBar(oIconTabBar);
        this.setObjectInspectorPopoverContainer(oObjectInspectorPopoverContainer);
        this.setButtonOverflowLayout(oButtonOverflowLayout);
        this.setButtonOverflowLayoutToolbar(oButtonOverflowLayoutToolbar);
        this.setOpenButton(oOpenButton);
        this.setEditButton(oEditButton);
        this.setFavoritesButton(oFavoritesButton);

        // give the icon tab bar to the details container
        sap.ui.getCore().byId(oObjectInspectorPopoverContainer.getDetailsContainer()).setMainContent(oIconTabBar);

        // add style-classes
        this.addStyleClass("sasObjectInspectorPopover");
        this.addStyleClass("oi_invisible"); // hide while sizing calculations/adjustments occur
        oOpenButton.addStyleClass('ButtonOverflowLayout_open');
        oEditButton.addStyleClass('ButtonOverflowLayout_edit');
        oEditButton.addStyleClass('ButtonOverflowLayout_favorites');

        // fake the control lifecycle b/c of ResponsivePopover extension
        this.attachBeforeOpen(this.preRenderingSteps);
        this.attachAfterOpen(this.postRenderingSteps);

        // set footer content depending on base class the responsive popover uses
        if (this._oControl instanceof sap.m.Popover) {
            this._oControl.setFooter(oButtonOverflowLayoutToolbar);
        } else {
            sas.log.warning("Setting footer on a sap.m.Dialog - faking it for now");
            this._oControl.addContent(oButtonOverflowLayoutToolbar);
        }

        // add main container class to responsive popover
        this.addContent(oObjectInspectorPopoverContainer);

        // keep around for legacy reasons: remove post-3.0
        this.oIconTabBar = oIconTabBar;
        this.oButtonOverflowLayout = oButtonOverflowLayout;
    };

    /**
     * This would be in the onBeforeRendering method, but there isn't really such a method
     * with responsive popovers.
     * Popup sizing, and simple table creation/updating
     * @private
     */
    ObjectInspectorBase.prototype.preRenderingSteps = function() {
        // objectData can be added at any order,
        // so sort according to spec before rendering
        this._sortObjectData();

        // create table from objectData
        this._manageSimpleTable();
    };

    /**
     * This would be in the onAfterRendering method, but there isn't really such a method
     * with responsive popovers.
     * @private
     */
    ObjectInspectorBase.prototype.postRenderingSteps = function() {
        var self = this;
        self._manageInitialFocus();
        setTimeout(function() {
            self._manageAutoSize();
        }, 100);
    };

    ObjectInspectorBase.prototype.destroy = function() {
        // If we do not pass in the suppress invalidate flag we get internal
        // errors later on when we try to actually call invalidate. Unclear why
        // this happens at all.
        ResponsivePopover.prototype.destroy.call(this, true);
    };


    // =============================================================================
    // SETTERS/GETTERS
    // =============================================================================

    ObjectInspectorBase.prototype.setHeaderText = function(sHeaderText) {
        this.setProperty("headerText", sHeaderText, true);
        sap.ui.getCore().byId(this.getHeaderTextLabel()).setText(sHeaderText);

        if (this.isOpen()) {
            this._refreshContent();
        }
    };
    ObjectInspectorBase.prototype.getHeaderText = function() {
        return sap.ui.getCore().byId(this.getHeaderTextLabel()).getText();
    };

    ObjectInspectorBase.prototype.setSecondaryText = function(sSecondaryText) {
        this.setProperty("secondaryText", sSecondaryText, true);
        sap.ui.getCore().byId(this.getSecondaryTextLabel()).setText(sSecondaryText);

        if (this.isOpen()) {
            this._refreshContent();
        }
    };
    ObjectInspectorBase.prototype.getSecondaryText = function() {
        return sap.ui.getCore().byId(this.getSecondaryTextLabel()).getText();
    };

    ObjectInspectorBase.prototype.setDescription = function(sDescription) {
        this.setProperty("description", sDescription, true);
        sap.ui.getCore().byId(this.getDescriptionLabel()).setText(sDescription);
    };
    ObjectInspectorBase.prototype.getDescription = function() {
        return sap.ui.getCore().byId(this.getDescriptionLabel()).getText();
    };

    ObjectInspectorBase.prototype.setIconKey = function(sIconKey) {
        var oHeaderIcon = sap.ui.getCore().byId(this.getHeaderIcon());
        this.setProperty("iconKey", sIconKey, true);
        oHeaderIcon.setSrc(sIconKey);
        oHeaderIcon.setVisible(!!sIconKey);

        if (this.isOpen()) {
            this._refreshContent();

        }
    };
    ObjectInspectorBase.prototype.getIconKey = function() {
        return sap.ui.getCore().byId(this.getHeaderIcon()).getSrc();
    };


    // =============================================================================
    // Manage Details "aggregation"
    // =============================================================================

    /**
     * detail aggregation addAggregation override
     * passes detail to IconTabBar
     * @param {sas.hc.m.IconTabFilter} oDetail The detail aggregation
     * @public
     */
    ObjectInspectorBase.prototype.addDetail = function(oDetail) {
        var oIconTabBar = sap.ui.getCore().byId(this.getIconTabBar());
        this.addAggregation("details", oDetail, true);
        oIconTabBar.addItem(oDetail);
    };

    /**
     * detail aggregation insertAggregation override
     * passes detail to IconTabBar
     * @param {sas.hc.m.IconTabFilter} oDetail The detail Aggregation
     * @param {integer} iIndex The index of the details aggregation to use
     * @public
     */
    ObjectInspectorBase.prototype.insertDetail = function(oDetail, iIndex) {
        var oIconTabBar = sap.ui.getCore().byId(this.getIconTabBar());
        this.insertAggregation("details", oDetail, iIndex, true);
        oIconTabBar.addItem(oDetail);
    };


    // =============================================================================
    // private functions
    // =============================================================================

    /**
     * Create, if not already created, a _SimplePairTable from objectData.
     * Otherwise update the existing simple table.
     * Manages simplePairTable association
     * @private
     */
    ObjectInspectorBase.prototype._manageSimpleTable = function() {
        var sId = this.getId(),
            oSimplePairTable = sap.ui.getCore().byId(this.getSimplePairTable()),
            aObjectData = this.getObjectData();

        // if table not already created,
        // create, add to IconTabBar, and store in an association
        // otherwise, assign content to table
        if (oSimplePairTable === undefined) {
            oSimplePairTable = new SimplePairTable({
                id: sId + "-SimplePairTable",
                data: aObjectData
            });
            this.insertDetail(new IconTabFilter({
                id: sId + "-IconTabFilter_simplePairTable",
                key: "defaultDataTable",
                text: this.rb.getText("objectInspector.details.txt"),
                content: [
                    oSimplePairTable
                ]
            }), 0);
            this.setSimplePairTable(oSimplePairTable);
        } else {
            oSimplePairTable.setData(aObjectData);
        }
    };

    /**
     * Add oData to objectData.
     * Depending on if oData's "name" property has already been added to objectData,
     * oData will be either appended, or will overwrite an existing same-name member.
     *
     * Since oData is a primitive array, it's listed as a property, not aggregation,
     * which means I need to roll-my-own "add" function.
     *
     * @param {object} oData  The data object to add
     * @private
     */
    ObjectInspectorBase.prototype._addObjectData = function(oData) {
        var aObjectData = this.getObjectData();

        if (!aObjectData) {
            aObjectData = [];
        }

        // might need to overwrite
        var bFound = false;
        for (var i = 0; i < aObjectData.length; i++) {
            var oObjectData = aObjectData[i];
            if (oObjectData.name === oData.name) {
                aObjectData[i].value = oData.value;
                bFound = true;
                break;
            }
        }

        if (bFound === false) {
            aObjectData.push(oData);
            this.setObjectData(aObjectData);
        }

        if (this.isOpen()) {
            this._updateSimplePairTable();
            this._refreshContent();
        }
    };

    /**
     * Remove opacity from ObjectInspector.
     * This adds a css class which uses animations to remove the opacity
     * This should occur after sizing calculations have taken place.
     * If animations have been turned off (for a11y) then simply remove
     * the css class that is applying the opacity
     * @private
     */
    ObjectInspectorBase.prototype._reveal = function() {
        var $this = this.$(),
            bUseAnimation = sas.hc.ui.core.settings !== undefined && sas.hc.ui.core.settings.GlobalSettingsHelper.getPlayVisualEffects() === true;
        if (bUseAnimation === true) {
            $this.addClass("oi_visible");
        } else {
            $this.removeClass("oi_invisible");
        }
    };

    /**
     * Set the contentHeight and contentWidth properties to match the actual content's height and width.
     * Try multiple times until the DOM has settled.
     * @param {int} iCount Number to determine how many times this function has been called (avoid infinite loops)
     * @private
     */
    ObjectInspectorBase.prototype._manageAutoSize = function(iCount) {
        var self = this,
            oObjectInspectorPopoverContainer = sap.ui.getCore().byId(this.getObjectInspectorPopoverContainer()),
            bDidChangeOccur = false,
            $this = this.$(),
            $body = $this.closest("body"),
            iMaxWidth = $body.outerWidth() / 2,
            $popoverContainer = oObjectInspectorPopoverContainer.$(),
            iDeterminedHeight, iDeterminedWidth;

        // skip auto-sizing
        if (this.getAutoHeight() === false || $popoverContainer.length <= 0) {
            this._reveal();
            return;
        }

        // initialize helper variables
        iCount = iCount || 0;
        this._previousHeight = this._previousHeight || 0;
        this._previousWidth = this._previousWidth || 0;

        // add 20 to account for scroll-bar
        // (assume scroll-bar will not have a height greater than 20)
        iDeterminedHeight = $popoverContainer.outerHeight() + 20;
        iDeterminedWidth = Math.min($popoverContainer.outerWidth(), iMaxWidth);

        // if height has changed
        if (iDeterminedHeight > this._previousHeight) {
            this.setContentHeight(iDeterminedHeight + "px");
            this._previousHeight = iDeterminedHeight;
            bDidChangeOccur = true;
        }

        // if width has changed
        if (iDeterminedWidth > this._previousWidth) {
            this.setContentWidth(iDeterminedWidth + "px");
            this._previousWidth = iDeterminedWidth;
            bDidChangeOccur = true;
        }

        // if change occurred, do another run to make sure DOM has settled
        // otherwise, reveal the properly-sized ObjectInspector
        if (bDidChangeOccur === true && iCount < 5) {
            setTimeout(function() {
                self._manageAutoSize(iCount+1);
            }, 100);
        } else {
            this._reveal();
        }
    };

    /**
     * Manage initial focus when the dialog appears
     * If there is a button in the bottom menu bar, give that focus,
     * Otherwise, just give the popover container focus.
     * @private
     */
    ObjectInspectorBase.prototype._manageInitialFocus = function() {
        var oDomRef = this.getDomRef();

        var $objectInspectorPopoverContainer = jQuery(".sasObjectInspectorPopoverContainer", oDomRef),
            $buttonOverflowLayout = jQuery(".ButtonOverflowLayout", oDomRef),
            domFirstButton = jQuery("button", $buttonOverflowLayout)[0];

        if (domFirstButton) {
            jQuery(domFirstButton).focus();
        } else {
            //$objectInspectorPopoverContainer.focus();
            //S1227049 
            oDomRef.focus();
        }
    };

    /**
     * Sort the objectData array so it is rendered in the correct order
     * Correct order (internal key in parens):
     *     - Type (objectType)
     *     - Location (objectLocation)
     *     - Date created (creationDate)
     *     - Date modified (modificationDate)
     *     - Keywords (keywords)
     *
     * @private
     */
    ObjectInspectorBase.prototype._sortObjectData = function() {
        var aSorted = [],
            aObjectData = this.getObjectData();

        // find sKey in aObjectData,
        // push it to aSorted,
        // remove the found object from aObjectData
        ["objectType","objectLocation","creationDate","modificationDate","keywords"].forEach(function(sKey) {
            var i = 0, oTargetObject;
            for (; i < aObjectData.length; i++) {
                oTargetObject = aObjectData[i];
                if (oTargetObject.key === sKey) {
                    aSorted.push(oTargetObject);
                    aObjectData.splice(i,1);
                    break;
                }
            }
        });

        // append the rest of the keys
        aSorted = aSorted.concat(aObjectData);

        this.setObjectData(aSorted);
    };

    ObjectInspectorBase.prototype._updateSimplePairTable = function() {
        this._sortObjectData();
        var oSimplePairTable = sap.ui.getCore().byId(this.getSimplePairTable());
        var aObjectData = this.getObjectData();
        oSimplePairTable.setData(aObjectData);
    };

    ObjectInspectorBase.prototype._refreshContent = function() {
        sap.ui.getCore().byId(this.getObjectInspectorPopoverContainer()).invalidate();

        var self = this;
        setTimeout(function() {
            self._manageAutoSize();
        }, 100);
    };

    return ObjectInspectorBase;
}, /* bExport= */ true);
