// (c) 2015, SAS Institute Inc.
sap.ui.define([
    "jquery.sap.global",
    "sap/ui/core/Control",
    "sas/hc/m/Button",
    "sas/hc/m/MenuButton",
    "sas/hc/m/Menu",
    "./_OIMenuItem"
], function(jQuery, Control, Button, MenuButton, Menu, OIMenuItem) {
    "use strict";

    /**
     * Container for action buttons in ObjectInspector.
     *
     * Manages rendering the "named" and "action" buttons.
     *
     * Button rendering approach via spec:
     * There are two classes of button: "named" and "action"
     * The named buttons are built-in, such as "Open", "Edit", and "Favorites"
     * The action buttons are added by the consumer.
     *
     * There are two rendering buckets these can go in:
     * The toolbar (here), and the overflow menu
     *
     * The named buttons get precedence...if there is room,
     * they are put on the toolbar.
     * If some of the named buttons are not desired by the consumer
     * (they set flags), then those empty slots get filled
     * by the action buttons if they exist.
     *
     * The remaining action buttons go to the overflow menu.
     * There are enough slots so that a named button will
     * never be rendered in the overflow menu.
     *
     * In LTR, the text-only buttons are left-aligned,
     * and the icon button are right-aligned. Reverse alignment for RTL.
     *
     * @class
     * @name sas.hc.m.objectInspector.ButtonOverflowLayout
     * @extends sap.ui.core.Control
     * @author Jonathan Brink
     * @version 207001.0.1.20170511082246_f0htmcm27p
     * @constructor
     * @private
     */
    var ButtonOverflowLayout = Control.extend("sas.hc.m.objectInspector.ButtonOverflowLayout", /** @lends sas.hc.m.objectInspector.ButtonOverflowLayout.prototype */ {
        metadata: {
            library: "sas.hc.m",
            aggregations: {
                menuButton: {type: "sas.hc.m.MenuButton", multiple: false}
            },
            associations: {
                objectInspector: {type: "sas.hc.m.objectInspector.ObjectInspector"},

                // belongs to MenuButton
                menu: {type: "sas.hc.ui.unified.Menu"}
            }
        },
        renderer: {
            render: function(oRenderManager, oControl) {
                var oObjectInspector = sap.ui.getCore().byId(oControl.getObjectInspector()),
                    aActionButtons = oObjectInspector.getActionButtons(),

                    // counter for the last action button to be rendered
                    iButtonIter = 0;

                oRenderManager.write('<div');
                oRenderManager.writeControlData(oControl);
                oRenderManager.addClass('ButtonOverflowLayout');
                oRenderManager.writeClasses();
                oRenderManager.write('>');

                // see the class comment at the top of this file for
                // details on this rendering logic

                // render named buttons: open and edit
                if (oObjectInspector.getShowOpenButton() === true) {
                    this._renderButton(oRenderManager, oObjectInspector.getOpenButton());
                }
                if (oObjectInspector.getShowEditButton() === true) {
                    this._renderButton(oRenderManager, oObjectInspector.getEditButton());
                }

                // render action buttons if there is room
                if (oObjectInspector.getShowOpenButton() !== true) {
                    this._renderButton(oRenderManager, aActionButtons[iButtonIter++]);
                }
                if (oObjectInspector.getShowEditButton() !== true) {
                    this._renderButton(oRenderManager, aActionButtons[iButtonIter++]);
                }

                // render named button: favorites
                if (oObjectInspector.getShowFavoritesButton() === true) {
                    this._renderButton(oRenderManager, oObjectInspector.getFavoritesButton());
                }

                // render action button if there is room
                if (oObjectInspector.getShowFavoritesButton() !== true) {
                    this._renderButton(oRenderManager, aActionButtons[iButtonIter++]);
                }

                // render the overflow button
                this._renderButton(oRenderManager, oControl.getMenuButton());

                oRenderManager.write('</div>');
            },

            _renderButton: function(oRenderManager, oButton) {
                if (oButton !== undefined) {
                    oRenderManager.renderControl(oButton);
                }
            }
        }
    });

    /**
     * init lifecycle callback
     * Creates actionSheet and overflowButton aggregations
     * @private
     */
    ButtonOverflowLayout.prototype.init = function() {
        var rb = sap.ui.getCore().getLibraryResourceBundle("sas.hc.m"),
            sId = this.getId(),

            // overflow menu
            oMenu = new Menu({
                id: sId + '-Menu',
                itemSelected: function(e) {
                    var oItem = e.getParameter('item'),
                        oAssociatedActionButton = sap.ui.getCore().byId(oItem.getButton());
                    oAssociatedActionButton.firePress();
                }
            }),

            // overflow button
            oMenuButton = new MenuButton({
                id: sId + '-MenuButton',
                visible: false,
                icon: sas.icons.HC.OVERFLOW,
                tooltip: rb.getText('objectInspector.moreOptions.txt'),
                menu: oMenu
            });

        oMenuButton.addStyleClass('ButtonOverflowLayout_Overflow');

        this.setMenuButton(oMenuButton);
        this.setMenu(oMenu);
    };

    /**
     * onBeforeRendering lifecycle callback
     * If overflow menu is empty, hide the overflow button
     * @private
     */
    ButtonOverflowLayout.prototype.onBeforeRendering = function() {
        var oMenu = sap.ui.getCore().byId(this.getMenu());
        oMenu.destroyItems();
    };

    /**
     * onAfterRendering lifecycle callback
     * If overflow menu is empty, hide the overflow button
     * @private
     */
    ButtonOverflowLayout.prototype.onAfterRendering = function() {
        var oObjectInspector = sap.ui.getCore().byId(this.getObjectInspector()),
            oMenu = sap.ui.getCore().byId(this.getMenu()),
            aActionButtons = oObjectInspector.getActionButtons();

        // populate overflow menu with remaining action buttons
        aActionButtons.forEach(function(oButton) {
            // skip is already rendered
            if (oButton.$() && oButton.$().length > 0) {
                return;
            }

            // port button characteristics to a MenuItem
            oMenu.addItem(new OIMenuItem({
                id: oMenu.getId() + "-menuItem" + oMenu.getItems().length,
                text: oButton.getText() || oButton.getTooltip_Text(),
                icon: oButton.getIcon(),
                button: oButton
            }));
        });

        // hide menu button is no items
        this.getMenuButton().setVisible(oMenu.getItems().length > 0);
    };

    // =============================================================================
    // LEGACY API - remove post-3.0
    // =============================================================================
    ButtonOverflowLayout.prototype.makeFavoriteButtonSelected = function() {
        var oObjectInspector = sap.ui.getCore().byId(this.getObjectInspector());
        oObjectInspector.getFavoritesButton().setSelected(true);
    };
    ButtonOverflowLayout.prototype.makeFavoriteButtonUnSelected = function() {
        var oObjectInspector = sap.ui.getCore().byId(this.getObjectInspector());
        oObjectInspector.getFavoritesButton().setSelected(false);
    };

    return ButtonOverflowLayout;
}, /* bExport= */ true);
